/*
  L-SHADE implemented by C++ for Special Session & Competition on Real-Parameter Single Objective Optimization at CEC-2014

  Version: 1.0   Date: 16/Apr/2014
  Written by Ryoji Tanabe (rt.ryoji.tanabe [at] gmail.com)
*/

//the code was changed by the authors of the paper: "Dynamic Perturbation for Population Diversity Management in Differential Evolution" to produce S_LSHADE_DP algorithm

//This code was downloaded from https://github.com/P-N-Suganthan/2022-SO-BO/tree/main in 2023. The code was changed by R. Biedrzycki rbiedrzy@elka.pw.edu.pl for the purpose of the paper: Revisiting CEC 2022 ranking for bound constrained single objective optimization.
//all numeric constants were named and their values can be specified when starting the program

#include <fstream>
#include <iomanip>
#include <sstream>

#include <boost/program_options.hpp>

#include "de.h"
namespace po = boost::program_options;

//g++ s_lshade_dp.cpp main.cpp search_algorithm.cpp cec22_test_func.cpp -lboost_program_options

//default:
//./a.out --seed 1234 --res_file_name default --pop_size 100 --mem_size 6 --arch_size_mult 2.6 --p_best_rate 0.11 --gamma 0.3 --upd_mut_op_iters 20 --min_pop_size 4 --cauchy_sigma 0.1 --initial_F 0.5 --stagnation_threshold 100 --CR_switch_at 0.5 --mut_switch_at 0.5 --stag_detect_at 0.5 --fit_epsilon 0 --loc_epsilon 0


#define NUM_RUN 30

double *OShift, *M, *y, *z, *x_bound;
int ini_flag = 0, n_flag, func_flag, *SS;

int g_function_number;
int g_problem_size;
unsigned int g_max_num_evaluations;
double g_optimum[12] = {300, 400, 600, 800, 900, 1800, 2000, 2200, 2300, 2400, 2600, 2700};

double fitness_memory[NUM_RUN][17];

void print_results(string file_name);

int main(int argc, char **argv){
  unsigned int seed=1234;
  unsigned int pop_size=100;
  unsigned int mem_size=6;
  double arch_size_mult=2.6;
  double p_best_rate = 0.11;
  double gamma = 0.3;
  unsigned int upd_mut_op_iters=20;
  int min_pop_size=4;
  double cauchy_sigma=0.1;
  double initial_F = 0.5;
  unsigned int stagnation_threshold = 100;
  double CR_switch_at=0.5;
  double mut_switch_at=0.5;
  double stag_detect_at=0.5;
  double fit_epsilon=0;
  double loc_epsilon=0;
  std::string res_file_name;

  po::options_description desc("Allowed options");
   desc.add_options ()
        ("help", "print usage message")
        ("seed", po::value(&seed), "seed")
        ("pop_size", po::value(&pop_size), "pop_size")
        ("mem_size", po::value(&mem_size), "memory size")
        ("arch_size_mult", po::value(&arch_size_mult), "archive size multiplier")
        ("p_best_rate", po::value(&p_best_rate), "p_best_rate")
        ("gamma", po::value(&gamma), "gamma")
        ("upd_mut_op_iters", po::value(&upd_mut_op_iters), "update mutation operator after iters")
        ("min_pop_size", po::value(&min_pop_size), "min_pop_size")
        ("cauchy_sigma", po::value(&cauchy_sigma), "cauchy_sigma")
        ("initial_F", po::value(&initial_F), "initial_F")
        ("stagnation_threshold", po::value(&stagnation_threshold), "stagnation_threshold")
        ("CR_switch_at", po::value(&CR_switch_at), "CR_switch_at")
        ("mut_switch_at", po::value(&mut_switch_at), "mut_switch_at")
        ("stag_detect_at", po::value(&stag_detect_at), "stag_detect_at")
        ("loc_epsilon", po::value(&loc_epsilon), "loc_epsilon")
        ("fit_epsilon", po::value(&fit_epsilon), "fit_epsilon")
        ("res_file_name", po::value<std::string>(&res_file_name)->required(), "result file base name")    
        ; 
    
    // Parse command line arguments
    po::variables_map vm;
    po::store (po::command_line_parser (argc, argv).options (desc).run (), vm);
    if(vm.count("help")){
        std::cout << desc << "\n";
        return 0;
    }
    po::notify(vm);
  loc_epsilon*=1e-8;
  fit_epsilon*=1e-8;
  srand(seed);
  
  //dimension size. please select from 10, 20
  for(int a_problem_size : {10,20}){
    g_problem_size=a_problem_size;

    //available number of fitness evaluations
    if (g_problem_size == 10)
    {
      g_max_num_evaluations = 200000;
    }
    else if (g_problem_size == 20)
    {
      g_max_num_evaluations = 1000000;
    }


    for (int i = 0; i < 12; i++)
    {
      g_function_number = i + 1;
      cout << "\n-------------------------------------------------------" << endl;
      cout << "Function = " << g_function_number << ", Dimension size = " << g_problem_size << "\n"
           << endl;

      Fitness *bsf_fitness_array = (Fitness *)malloc(sizeof(Fitness) * NUM_RUN);
      Fitness mean_bsf_fitness = 0;
      Fitness std_bsf_fitness = 0;

      for (int j = 0; j < NUM_RUN; j++)
      {
        //random seed is selected based on time according to competition rules
        //int seed_ind = (g_problem_size / 10 * g_function_number * NUM_RUN + j) - NUM_RUN;
        //seed_ind %= 1000;
        //srand(random_seeds[seed_ind]);
        
        string startPopFilename="../startPointsGen/startingPoints";
        startPopFilename+=to_string(a_problem_size)+'_'+to_string(j)+'_'+to_string(g_function_number)+".txt";
        ifstream fin(startPopFilename.c_str());
        vector<vector<double>> startPop;
        if(fin.is_open()){
          double f;
          char c;
          for(int numOfStartPoints = 0;numOfStartPoints<pop_size; ++numOfStartPoints){
            vector<double> startInd;
            for(int curDim = 0;curDim<a_problem_size; ++curDim){
              fin>>f;
              fin>>c;
              startInd.push_back(f);
            }
            startPop.emplace_back(startInd);
          }
        }else{
          cout<<startPopFilename<<" open failed"<<endl;
          return -1;
        }
                    

        searchAlgorithm *alg = new S_LSHADE_DP();
        bsf_fitness_array[j] = alg->run(&fitness_memory[j][0], pop_size, mem_size, arch_size_mult, p_best_rate, gamma, upd_mut_op_iters, min_pop_size,
                                        cauchy_sigma, initial_F,stagnation_threshold, CR_switch_at, mut_switch_at, stag_detect_at, loc_epsilon, fit_epsilon, startPop);
        cout << j + 1 << "th run, "
             << "error value = " << bsf_fitness_array[j] << endl;
      }

      // print results to file
       stringstream file_name;
       
       file_name << res_file_name+"_" << g_function_number << "_" << g_problem_size << ".txt";
       print_results(file_name.str());

      for (int j = 0; j < NUM_RUN; j++)
        mean_bsf_fitness += bsf_fitness_array[j];
      mean_bsf_fitness /= NUM_RUN;

      for (int j = 0; j < NUM_RUN; j++)
        std_bsf_fitness += pow((mean_bsf_fitness - bsf_fitness_array[j]), 2.0);
      std_bsf_fitness /= NUM_RUN;
      std_bsf_fitness = sqrt(std_bsf_fitness);

      cout << "\nmean = " << mean_bsf_fitness << ", std = " << std_bsf_fitness << endl;
      free(bsf_fitness_array);
    }
  }
  return 0;
}


/*void read_random_seed(double *seed_list, int seed_num)
{
  string file_path = "input_data/Rand_Seeds.txt";
  ifstream fin(file_path);

  if (fin.is_open())
  {
    for (int i = 0; i < seed_num; i++)
    {
      fin >> seed_list[i];
    }

    cout << "Random seeds imported successfully!" << endl;
  }
  else
  {
    cout << "Error! cannot onpen random_seeds file." << endl;
  }

  fin.close();
}*/

void print_results(string file_path){
  ofstream fout(file_path);
  if (fout.is_open())
  {
    fout << scientific << setprecision(6);
    for (int i = 0; i < 17; i++)
    {
      for (int j = 0; j < NUM_RUN; j++)
      {
        fout << fitness_memory[j][i] << " ";
      }
      fout << endl;
    }
  }
  else
  {
    cout << "Error! cannot onpen output file " << file_path << "." << endl;
  }

  fout.close();
}
