# -*- coding: utf-8 -*-
# This is the code that calculates ranking and percentage quality measure of optimization algorithms, which was described in:
# Rafał Biedrzycki: Revisiting CEC 2022 ranking: A new ranking method and influence of parameter tuning
# https://www.sciencedirect.com/science/article/pii/S2210650224001615
# author: Rafał Biedrzycki
# license: MIT
# input data format: https://github.com/P-N-Suganthan/2022-SO-BO/blob/main/CEC2022%20TR.pdf
# According to the rules of the CEC 2022 SOBC competition, all values below 1e-8 are set here to 1e-8

# As I believe that:
# 'A difference is a difference only if it makes a difference.'
#  - Darrell Huff, How to Lie with Statistics 
# the error values are rounded to 5 significant digits, and the FES (objective functions evaluations) are rounded to tenths

# exemplary results to analyze
alg_names = ["simplifiedTuned/onlyIDEcorrItMaxWithoutSortsNoIDEbdmutNoRandxr3_3per",
            "CEC2022_EA4eig_CORRECTED/CEC2022_EA4eig_CORRECTED"]
NUM_OF_ALGS = len(alg_names)

import os.path
import sys
import pandas as pd
import math
import numpy as np

#singnif from https://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python
def signif(x, digits=6):
    if x == 0 or not math.isfinite(x):
        return x
    digits -= math.ceil(math.log10(abs(x)))
    return round(x, digits)

vsignif = np.vectorize(signif)

MAX_FUN_NMBR = 12
NEXT_LEVEL_WEIGHT = 0.01 #weight for less important criterion

SINIF_DIGITS = 5 #for rounding error. signif(12345,3)==12300; signif(0.000000012345,3)==0.0000000123
SINIF_DIGITS4FES = -1 #for rounding FES. round(12345, -1)==12340;

dims= ["10", "20"]
STOP_FIT=1e-8
max_FES = 0
max_FES_tab = [2e5, 1e6]
NMBR_OF_RUNS=30

#create thresholds
NUM_OF_THRES=51
thres_seq_start = 3.0
thres_seq_stop = -8.0

thres_exponents = np.linspace( thres_seq_start, thres_seq_stop, num=NUM_OF_THRES)
thresholds = np.power(10.0, thres_exponents) 

glob_algs_marks = {}
glob_algs_marks_2_displ = {}
algs_marks_4_dim= {"10":{},"20":{}}
algs_marks_4_dim_2_displ= {"10":{},"20":{}}

algs_marks_4_fun= {"10":{"1":{}}, "20":{"1":{}}}
algs_marks_4_fun_2_displ= {"10":{"1":{}}, "20":{"1":{}}}

#prepare dict 
for dim in dims:
    for fun_number in range(1, MAX_FUN_NMBR+1):          
        algs_marks_4_fun[dim][str(fun_number)]={}
        algs_marks_4_fun_2_displ[dim][str(fun_number)]={}
            
for ALG_NR in range(0,NUM_OF_ALGS):
    res_file_base_name=alg_names[ALG_NR] 

    num_of_glob_opt_found_4_alg = 0
    num_of_thres_achieved_4_alg = 0
    num_of_used_evals_4_alg = 0

    for dim in dims:
        if dim=='10':
            max_FES = max_FES_tab[0]
        else:
            max_FES = max_FES_tab[1]
            
        num_of_glob_opt_found_4_dim = 0
        num_of_thres_achieved_4_dim = 0
        num_of_used_evals_4_dim = 0        
        for fun_number in range(1, MAX_FUN_NMBR+1):

            results_file_name = res_file_base_name + "_"+ str(fun_number) +"_"+dim+ ".txt"

            if not os.path.isfile(results_file_name):
                print( results_file_name  + " not found.")
        
            alg_results = pd.read_table(results_file_name, header=None, index_col=False, sep='\\s+')

            stoped_at_evals = round(alg_results.iloc[16,:], SINIF_DIGITS4FES) 
            stoped_at_evals[stoped_at_evals>max_FES] = max_FES
            
            error_of_best_sols = vsignif(alg_results.iloc[15,:], SINIF_DIGITS)
            error_of_best_sols[error_of_best_sols<1e-8] = 1e-8
            num_of_glob_opt_found_4_fun = sum(error_of_best_sols <= 1e-8)
            per_of_glob_opt_found_4_fun = 100.0 * num_of_glob_opt_found_4_fun/NMBR_OF_RUNS  
            
            num_of_thres_achieved_4_fun = 0
            for thr in thresholds:
                num_of_thres_achieved_4_fun += sum(error_of_best_sols <= thr)
            per_of_thres_achieved_4_fun = 100.0 * num_of_thres_achieved_4_fun/(NMBR_OF_RUNS*NUM_OF_THRES)
            
            num_of_used_evals_4_fun = sum(stoped_at_evals)
            per_of_free_evals_4_fun = 100.0 - 100.0 * num_of_used_evals_4_fun/(max_FES*NMBR_OF_RUNS)
            
            weighted_res = per_of_glob_opt_found_4_fun+per_of_thres_achieved_4_fun*NEXT_LEVEL_WEIGHT+per_of_free_evals_4_fun*NEXT_LEVEL_WEIGHT*NEXT_LEVEL_WEIGHT
            num_of_glob_opt_found_4_dim = num_of_glob_opt_found_4_dim + num_of_glob_opt_found_4_fun
            num_of_thres_achieved_4_dim = num_of_thres_achieved_4_dim + num_of_thres_achieved_4_fun
            num_of_used_evals_4_dim = num_of_used_evals_4_dim + num_of_used_evals_4_fun
            algs_marks_4_fun[dim][str(fun_number)][res_file_base_name] = weighted_res
            algs_marks_4_fun_2_displ[dim][str(fun_number)][res_file_base_name] = "{} {} {}".format(round(per_of_glob_opt_found_4_fun), round(per_of_thres_achieved_4_fun), round(per_of_free_evals_4_fun))

        res_4_dim = num_of_glob_opt_found_4_dim/(NMBR_OF_RUNS*MAX_FUN_NMBR)*100
        per_of_achieved_thres_4_dim = num_of_thres_achieved_4_dim/(NMBR_OF_RUNS*MAX_FUN_NMBR*len(thresholds))*100
        res_4_dim = res_4_dim+per_of_achieved_thres_4_dim*NEXT_LEVEL_WEIGHT
        per_of_free_evals_4_dim = 100-num_of_used_evals_4_dim/(max_FES*NMBR_OF_RUNS*MAX_FUN_NMBR)*100
        res_4_dim = res_4_dim+per_of_free_evals_4_dim*NEXT_LEVEL_WEIGHT*NEXT_LEVEL_WEIGHT
        
        num_of_glob_opt_found_4_alg = num_of_glob_opt_found_4_alg + num_of_glob_opt_found_4_dim
        num_of_thres_achieved_4_alg = num_of_thres_achieved_4_alg + num_of_thres_achieved_4_dim
        num_of_used_evals_4_alg = num_of_used_evals_4_alg + num_of_used_evals_4_dim
        algs_marks_4_dim[dim][res_file_base_name] = res_4_dim
        algs_marks_4_dim_2_displ[dim][res_file_base_name] = "{} {} {}".format(round(num_of_glob_opt_found_4_dim/(NMBR_OF_RUNS*MAX_FUN_NMBR)*100), round(per_of_achieved_thres_4_dim), round(per_of_free_evals_4_dim))
            
            
    res_4_alg = num_of_glob_opt_found_4_alg/(NMBR_OF_RUNS*MAX_FUN_NMBR*len(dims))*100
    per_of_achieved_thres_4_alg = num_of_thres_achieved_4_alg/(NMBR_OF_RUNS*MAX_FUN_NMBR*len(dims)*len(thresholds))*100
    res_4_alg = res_4_alg+per_of_achieved_thres_4_alg*NEXT_LEVEL_WEIGHT
    per_of_free_evals_4_alg = 100-num_of_used_evals_4_alg/((max_FES_tab[0]+max_FES_tab[1])*NMBR_OF_RUNS*MAX_FUN_NMBR)*100
    res_4_alg = res_4_alg+per_of_free_evals_4_alg*NEXT_LEVEL_WEIGHT*NEXT_LEVEL_WEIGHT
    glob_algs_marks[res_file_base_name] = res_4_alg
    glob_algs_marks_2_displ[res_file_base_name] = "{} {} {}".format(round(num_of_glob_opt_found_4_alg/(NMBR_OF_RUNS*MAX_FUN_NMBR*len(dims))*100), round(per_of_achieved_thres_4_alg), round(per_of_free_evals_4_alg))
          
print("Results for each function")
for dim in dims:
    print(dim, "D:")
    for fun_number in range(1, MAX_FUN_NMBR+1):
        print("fun:", fun_number)
        alg_marks_4_dim_4_fun = algs_marks_4_fun[dim][str(fun_number)]
        alg_marks_4_dim_4_fun_2_displ = algs_marks_4_fun_2_displ[dim][str(fun_number)]
        d_view = [ (v,k) for k,v in alg_marks_4_dim_4_fun.items() ]
        d_view.sort(reverse=True)
        for v, k in d_view:
            print( "%s: %s" % (k, alg_marks_4_dim_4_fun_2_displ[k]) )

print("Results for each dimensionality")
for dim in dims:
    print(dim, "D:")
    alg_marks_4_curr_dim = algs_marks_4_dim[dim]
    alg_marks_4_curr_dim_2_displ = algs_marks_4_dim_2_displ[dim]
    d_view = [ (v,k) for k,v in alg_marks_4_curr_dim.items() ]
    d_view.sort(reverse=True)
    for v, k in d_view:
        print( "%s: %s" % (k, alg_marks_4_curr_dim_2_displ[k]) )
  
print("\n-----------------------------------------------")
print("Final, aggregated results")
d_view = [ (v,k) for k,v in glob_algs_marks.items() ]
d_view.sort(reverse=True)
r=1
for v, k in d_view:
    print( "%d %s: %s" % (r, k, glob_algs_marks_2_displ[k]) )
    r+=1


    

    
