// This is the code of nl_shade_rsp_mid.
// Most of the code is taken from nl_shade_rsp, which was authored by:
// Vladimir Stanovov, Shakhnaz  Akhmedova, Eugene Semenkin 
// and downloaded from: https://github.com/P-N-Suganthan
// Changes were conceptually described in the article: "A Version of NL-SHADE-RSP Algorithm with Midpoint for CEC 2022 Single Objective Bound Constrained Problems"
// Author of the changes: Rafal Biedrzycki

//Modifications in 2023 for the paper "Revisiting CEC 2022 ranking for bound constrained single objective optimization":
//1) all numeric constants were named now and their values can be specified when starting the program

// To compile: (tested under Ubuntu 22.04)
// Use system package manager to install libraries: mlpack and boost (libmlpack-dev, libboost-all-dev)
// compile & link by command: g++ -std=c++17 nl_shade_rsp_mid.cpp -lmlpack -fopenmp -lboost_program_options

// Before running the program, run command (or-else program will use many threads wchich slowes down in my case):
// export OMP_NUM_THREADS=1
//default              ./a.out --seed 1234 --res_file_name default --pop_size_mult 5 --k_means 1 --resampling 1 --pop_size_reset 400 --shape_const 0.1  --min_pop_size_aft_restart 20 --count_lim 1 --min_num_stag_it 8 --min_its_on_bounds 9 --min_silhouette_mult 0.025

#include <cmath>
#include <iostream>
#include <fstream>
#include <random>

#include <boost/program_options.hpp>
#include <boost/math/statistics/linear_regression.hpp>
#include <cstring>
#include <mlpack/methods/kmeans/kmeans.hpp>
#include <mlpack/core/cv/metrics/silhouette_score.hpp>

//the code of test functions to be optimized
#include "RB_cec22_test_func.cpp"

using namespace mlpack::kmeans;
using namespace mlpack::cv;
namespace po = boost::program_options;

using namespace std;
using boost::math::statistics::simple_ordinary_least_squares_with_R_squared;



std::mt19937 generator;
std::uniform_int_distribution<int> uni_int(0,32768);
std::uniform_real_distribution<double> uni_real(0.0,1.0);
std::normal_distribution<double> norm_dist(0.0,1.0);
std::cauchy_distribution<double> cachy_dist(0.0,1.0);

const double MIN_ERROR=1e-8;

int intRandom(int target);
double Random(double minimal, double maximal){return uni_real(generator)*(maximal-minimal)+minimal;}
double NormRand(double mu, double sigma){return norm_dist(generator)*sigma + mu;}
double CachyRand(double mu, double sigma){return cachy_dist(generator)*sigma+mu;}

void getMean(double** popul, int popul_size, int dimensionality, double* mean_indiv){
  for(int dim=0;dim<dimensionality;++dim){
    double sum=0;
    for(int indIndx=0;indIndx<popul_size;++indIndx){
      sum+=popul[indIndx][dim];
    }
    mean_indiv[dim] = sum/popul_size;
  }
}

int getIndxOfNearest(double** popul, int popul_size, int dimensionality, double* mean_indiv){
    int nearestIndx=0;
    double minDist = 1e20;
    for(int indIndx=0;indIndx<popul_size;++indIndx){
        double sqDist=0;
        for(int dim=0;dim<dimensionality;++dim){
            sqDist+= pow(mean_indiv[dim]-popul[indIndx][dim],2);
        }
        if(sqDist<minDist){
            minDist=sqDist;
            nearestIndx = indIndx;
        }
    }
    return nearestIndx;
}

double getDist(double* ind1, double* ind2, int dimensionality){
    double dist=0.0;
    for(int dim=0;dim<dimensionality;++dim){
        dist+= pow(ind1[dim]-ind2[dim], 2);
    }
    return sqrt(dist);
}

void qSort1(double* mass, int low, int high)
{
    int i=low;
    int j=high;
    double x=mass[(low+high)>>1];
    do
    {
        while(mass[i]<x)    ++i;
        while(mass[j]>x)    --j;
        if(i<=j)
        {
            double temp=mass[i];
            mass[i]=mass[j];
            mass[j]=temp;
            i++;    j--;
        }
    } while(i<=j);
    if(low<j)   qSort1(mass,low,j);
    if(i<high)  qSort1(mass,i,high);
}

void qSort2int(double* mass, int* mass2, int low, int high)
{
    int i=low;
    int j=high;
    double x=mass[(low+high)>>1];
    do
    {
        while(mass[i]<x)    ++i;
        while(mass[j]>x)    --j;
        if(i<=j)
        {
            double temp=mass[i];
            mass[i]=mass[j];
            mass[j]=temp;
            int temp2=mass2[i];
            mass2[i]=mass2[j];
            mass2[j]=temp2;
            i++;    j--;
        }
    } while(i<=j);
    if(low<j)
        qSort2int(mass,mass2,low,j);
    if(i<high)
        qSort2int(mass,mass2,i,high);
}

void cec22_test_func(double *, double *,int,int,int);
double *OShift,*M,*y,*z,*x_bound;
int ini_flag=0,n_flag,func_flag,*SS;

//const int dimensionality=10;
//const int dimensionality=20;

const int NMBR_OF_RUNS=30;
const int NUM_OF_DUMPS=16;
double stepsFEval[NUM_OF_DUMPS];
//double ResultsArray[NUM_OF_DUMPS+1];
double ResultsArray[NMBR_OF_RUNS][NUM_OF_DUMPS+1]={0};
int LastFEcount;
int NFEval = 0;
int maxFES = 0;
int evalsAtStart=0;
double tempF[1];
double fopt;
char buffer[200];
double globalBestFit;
bool globalBestFitinit;
bool isRestart=false;

//double bestSol[dimensionality];
bool initfinished;
vector<double> FitTemp3;
int foundAt=-1;

void GenerateNextRandUnif(const int num, const int Range, int* Rands, const int Prohib)
{
    for(int j=0;j!=25;j++)
    {
        bool generateagain = false;
        Rands[num] = intRandom(Range);
        for(int i=0;i!=num;i++)
            if(Rands[i] == Rands[num])
                generateagain = true;
        if(!generateagain)
            break;
    }
}

void GenerateNextRandUnifOnlyArch(const int num, const int Range, const int Range2, int* Rands, const int Prohib)
{
    for(int j=0;j!=25;j++)
    {
        bool generateagain = false;
        Rands[num] = intRandom(Range2)+Range;
        for(int i=0;i!=num;i++)
            if(Rands[i] == Rands[num])
                generateagain = true;
        if(!generateagain)
            break;
    }
}

bool CheckGenerated(const int num, int* Rands, const int Prohib)
{
    if(Rands[num] == Prohib)
        return false;
    for(int j=0;j!=num;j++)
        if(Rands[j] == Rands[num])
            return false;
    return true;
}

void SaveBestValues(int funcN, int runNum)
{
    for(int stepFEcount=LastFEcount;stepFEcount<NUM_OF_DUMPS;stepFEcount++)
    {
        if(NFEval == int(stepsFEval[stepFEcount]*maxFES))
        {
            double temp = globalBestFit - fopt;
            if(temp <= MIN_ERROR)
                temp = 0;
            ResultsArray[runNum][stepFEcount] = temp;
            LastFEcount = stepFEcount;
        }
    }
    
}

void FindLimits(double* Ind, double* Parent, int CurNVars, double CurLeft, double CurRight){
    for (int j = 0; j<CurNVars; ++j){
        if (Ind[j] < CurLeft)
            Ind[j] = Random(CurLeft,CurRight);
        if (Ind[j] > CurRight)
            Ind[j] = Random(CurLeft,CurRight);
    }
}

void countLimits(double* ind, int* indLimCounter, int dimensionality, double lower, double upper){
    bool onBound=false;
    for (int dim = 0; dim<dimensionality; ++dim){
        if (ind[dim] <= lower){
            onBound=true;
            break;
        }
        if (ind[dim] >= upper){
            onBound=true;
            break;
        }
    }
    if(onBound){
        ++(*indLimCounter);
    }else{
        *indLimCounter=0;
    }
}

bool IsInfeasible(double* Ind, int CurNVars,double CurLeft, double CurRight){
    for (int j = 0; j<CurNVars; ++j){
        if (Ind[j] < CurLeft)
            return true;
        if (Ind[j] > CurRight)
            return true;
    }
    return false;
}

class Optimizer
{
public:
    int runNum;
    int dim;
    //double bestSol[dim];
    int k_means;
    int resampling;
    double shape_const;
    int min_pop_size_aft_restart;
    int count_lim;
    int min_num_stag_it;
    double min_silhouette_mult;
    int min_its_on_bounds;

    double initial_CR;
    double initial_F;
    int min_pop_size;
    double arch_prob;
    double prob_use_exp;
    double when_start_2_use_CRtoUse;
    double per_of_the_best_considered_at_start;
    double per_of_the_best_considered_at_end;
    double norm_sigma;
    double cauchy_sigma;
    double default_CR;
    double default_F;    


    bool FitNotCalculated;
    int Int_ArchiveSizeParam;
    int MemorySize;
    int MemoryIter;
    int SuccessFilled;
    int MemoryCurrentIndex;
    int NVars;
    int popSize;
    int NIndsMax;
    int NIndsMin;
    int besti;
    int func_num;
    int Generation;
    int ArchiveSize;
    int CurrentArchiveSize;
    double F;
    double Cr;
    double bestfit;
    double ArchiveSizeParam;
    double Right;
    double Left;
    int* Rands;
    int* Indexes;
    int* BackIndexes;
    double* Weights;
    double* Donor;
    double* Trial;
    double* Fitmass;
    double* popFitTmp;
    double* FitmassCopy;
    double* BestInd;
    double* tempSuccessCr;
    double* tempSuccessF;
    double* FGenerated;
    double* CrGenerated;
    double* MemoryCr;
    double* MemoryF;
    double* FitDelta;
    double* ArchUsages;
    double** Popul;
    double** populTemp;
    double** Archive;
    ofstream infoLog;

    void Initialize(int newNInds, int newNVars, int new_func_num,
                           int NewMemSize, double NewArchSizeParam, int k_means, int resampling, double shape_const, int min_pop_size_aft_restart, int count_lim, int min_num_stag_it, double min_silhouette_mult,
                           int min_its_on_bounds,
                           float initial_CR, float initial_F, int min_pop_size, float arch_prob, float prob_use_exp, float when_start_2_use_CRtoUse, 
                           float per_of_the_best_considered_at_start, float per_of_the_best_considered_at_end, 
                           float norm_sigma, float cauchy_sigma, float default_CR, float default_F, int runNumm, const vector<vector<double>>& startPop);

    void restart(int newPopSize);
    void Clean();
    void MainCycle(double);
    void FindNSaveBest(bool init, int ChosenOne);
    inline double GetValue(const int index, const int popSize, const int j);
    void CopyToArchive(double* RefusedParent);
    void SaveSuccessCrF(double Cr, double F, double FitD);
    void UpdateMemoryCrF();
    double MeanWL_general(double* Vector, double* TempWeights, int Size, double g_p, double g_m);
    void RemoveWorst(int popSize, int NewNInds);
    void RemoveTooNear(int popSize, int NewNInds);
};

double cec_22_(double* HostVector,int func_num, int dim){
	int nx = dim;	
	int mx = 1;
	
	cec22_test_func(HostVector, tempF, nx, mx,func_num);
    NFEval++;
    return tempF[0];
}

void Optimizer::Initialize(int newNInds, int newNVars, int new_func_num,
                           int NewMemSize, double NewArchSizeParam, int k_means, int resampling, double shape_const, int min_pop_size_aft_restart, int count_lim, int min_num_stag_it, double min_silhouette_mult,
                           int min_its_on_bounds,
                           float initial_CR, float initial_F, int min_pop_size, float arch_prob, float prob_use_exp, float when_start_2_use_CRtoUse, 
                           float per_of_the_best_considered_at_start, float per_of_the_best_considered_at_end, 
                           float norm_sigma, float cauchy_sigma, float default_CR, float default_F, int runNum, const vector<vector<double>>& startPop
)
{
    this->runNum = runNum;
    this->dim = newNVars;
    this->k_means=k_means;
    this->resampling=resampling;
    this->shape_const=shape_const;
    this->min_pop_size_aft_restart=min_pop_size_aft_restart;
    this->count_lim=count_lim;
    this->min_num_stag_it=min_num_stag_it;
    this->min_silhouette_mult=min_silhouette_mult;
    this->min_its_on_bounds=min_its_on_bounds;

    this->initial_CR=initial_CR;
    this->initial_F=initial_F;
    this->min_pop_size=min_pop_size;
    this->arch_prob=arch_prob;
    this->prob_use_exp=prob_use_exp;
    this->when_start_2_use_CRtoUse=when_start_2_use_CRtoUse;
    this->per_of_the_best_considered_at_start=per_of_the_best_considered_at_start;
    this->per_of_the_best_considered_at_end= per_of_the_best_considered_at_end;
    this->norm_sigma=norm_sigma;
    this->cauchy_sigma=cauchy_sigma;
    this->default_CR=default_CR;
    this->default_F=default_F;

    for(int j=0;j<=NUM_OF_DUMPS;++j)
        ResultsArray[runNum][j]=0.0;

    evalsAtStart=0;
    FitNotCalculated = true;
    popSize = newNInds;
    NIndsMax = popSize;
    NIndsMin = min_pop_size;//originalny rozm populacji
    NVars = newNVars;
    Left = -100;//+10000;
    Right = 100;//+10000;
    Cr = initial_CR;
    F = initial_F;
    besti = 0;
    Generation = 0;
    CurrentArchiveSize = 0;
    ArchiveSizeParam = NewArchSizeParam;
    Int_ArchiveSizeParam = ceil(ArchiveSizeParam);
    ArchiveSize = NIndsMax*ArchiveSizeParam;
    func_num = new_func_num;

    for(int steps_k=0;steps_k!=NUM_OF_DUMPS-1;steps_k++)
        stepsFEval[steps_k] = pow(double(dim),double(steps_k)/5.0-3.0);
    stepsFEval[NUM_OF_DUMPS-1] = 1.0;
    Popul = new double*[NIndsMax];
    for(int i=0;i!=NIndsMax;i++)
        Popul[i] = new double[NVars];
    populTemp = new double*[NIndsMax];
    for(int i=0;i!=NIndsMax;i++)
        populTemp[i] = new double[NVars];
    Archive = new double*[NIndsMax*Int_ArchiveSizeParam];
    for(int i=0;i!=NIndsMax*Int_ArchiveSizeParam;i++)
        Archive[i] = new double[NVars];
    Fitmass = new double[NIndsMax];
    popFitTmp = new double[NIndsMax];
    FitmassCopy = new double[NIndsMax];
    Indexes = new int[NIndsMax];
    BackIndexes = new int[NIndsMax];
    BestInd = new double[NVars];
	for (int i = 0; i<NIndsMax; i++)
		for (int j = 0; j<NVars; j++)
			Popul[i][j] = startPop[i][j];//Random(Left,Right);
    Donor = new double[NVars];
    Trial = new double[NVars];
    Rands = new int[NIndsMax];
    tempSuccessCr = new double[NIndsMax];
    tempSuccessF = new double[NIndsMax];
    FitDelta = new double[NIndsMax];
    FGenerated = new double[NIndsMax];
    CrGenerated = new double[NIndsMax];
    for(int i=0;i!=NIndsMax;i++)
    {
        tempSuccessCr[i] = 0;
        tempSuccessF[i] = 0;
    }
    MemorySize = NewMemSize;
    MemoryIter = 0;
    SuccessFilled = 0;
    ArchUsages = new double[NIndsMax];
    Weights = new double[NIndsMax];
    MemoryCr = new double[MemorySize];
    MemoryF = new double[MemorySize];
    for(int i=0;i!=MemorySize;i++)
    {
        MemoryCr[i] = initial_CR;
        MemoryF[i] = initial_F;
    }
    foundAt=maxFES;

}

void Optimizer::restart(int newPopSize){
    evalsAtStart = NFEval;
    FitNotCalculated = true;
    popSize = newPopSize;
    NIndsMax = popSize;
    NIndsMin = min_pop_size_aft_restart;
    Cr = initial_CR;
    F =initial_F;
    besti = 0;
    Generation = 0;
    CurrentArchiveSize = 0;
    ArchiveSize = NIndsMax*ArchiveSizeParam;

    Popul = new double*[NIndsMax];
    for(int i=0;i!=NIndsMax;i++)
        Popul[i] = new double[NVars];
    populTemp = new double*[NIndsMax];
    for(int i=0;i!=NIndsMax;i++)
        populTemp[i] = new double[NVars];
    Archive = new double*[NIndsMax*Int_ArchiveSizeParam];
    for(int i=0;i!=NIndsMax*Int_ArchiveSizeParam;i++)
        Archive[i] = new double[NVars];
    Fitmass = new double[NIndsMax];
    popFitTmp = new double[NIndsMax];
    FitmassCopy = new double[NIndsMax];
    Indexes = new int[NIndsMax];
    BackIndexes = new int[NIndsMax];
    BestInd = new double[NVars];

    
	for (int i = 0; i<NIndsMax; i++)
		for (int j = 0; j<NVars; j++)
			Popul[i][j] = Random(Left,Right);
            


    Rands = new int[NIndsMax];
    tempSuccessCr = new double[NIndsMax];
    tempSuccessF = new double[NIndsMax];
    FitDelta = new double[NIndsMax];
    FGenerated = new double[NIndsMax];
    CrGenerated = new double[NIndsMax];
    for(int i=0;i!=NIndsMax;i++)
    {
        tempSuccessCr[i] = 0;
        tempSuccessF[i] = 0;
    }
    MemoryIter = 0;
    SuccessFilled = 0;
    ArchUsages = new double[NIndsMax];
    Weights = new double[NIndsMax];
    MemoryCr = new double[MemorySize];
    MemoryF = new double[MemorySize];
    for(int i=0;i!=MemorySize;i++)
    {
        MemoryCr[i] = initial_CR;
        MemoryF[i] = initial_F;
    }


}

void Optimizer::SaveSuccessCrF(double Cr, double F, double FitD)
{
    tempSuccessCr[SuccessFilled] = Cr;
    tempSuccessF[SuccessFilled] = F;
    FitDelta[SuccessFilled] = FitD;
    SuccessFilled++;
}
void Optimizer::UpdateMemoryCrF()
{
    if(SuccessFilled != 0)
    {
        MemoryCr[MemoryIter] =MeanWL_general(tempSuccessCr,FitDelta,SuccessFilled,2,1);
        MemoryF[MemoryIter] = MeanWL_general(tempSuccessF, FitDelta,SuccessFilled,2,1);
        MemoryIter++;
        if(MemoryIter >= MemorySize)
            MemoryIter = 0;
    }
    else
    {
        MemoryF[MemoryIter] = default_F;
        MemoryCr[MemoryIter] = default_CR;
    }
}
double Optimizer::MeanWL_general(double* Vector, double* TempWeights, int Size, double g_p, double g_m)
{
    double SumWeight = 0;
    double SumSquare = 0;
    double Sum = 0;
    for(int i=0;i!=SuccessFilled;i++)
        SumWeight += TempWeights[i];
    for(int i=0;i!=SuccessFilled;i++)
        Weights[i] = TempWeights[i]/SumWeight;
    for(int i=0;i!=SuccessFilled;i++)
        SumSquare += Weights[i]*pow(Vector[i],g_p);
    for(int i=0;i!=SuccessFilled;i++)
        Sum += Weights[i]*pow(Vector[i],g_p-g_m);
    if(fabs(Sum) > 0.000001)
        return SumSquare/Sum;
    else
        return 0.5;
}

void Optimizer::CopyToArchive(double* RefusedParent)
{
    if(CurrentArchiveSize < ArchiveSize)
    {
        for(int i=0;i!=NVars;i++)
            Archive[CurrentArchiveSize][i] = RefusedParent[i];
        CurrentArchiveSize++;
    }
    else if(ArchiveSize > 0)
    {
        int RandomNum = intRandom(ArchiveSize);
        for(int i=0;i!=NVars;i++)
            Archive[RandomNum][i] = RefusedParent[i];
    }
}

void Optimizer::FindNSaveBest(bool init, int ChosenOne)
{
    if(Fitmass[ChosenOne] <= bestfit || init)
    {
        bestfit = Fitmass[ChosenOne];
        besti = ChosenOne;
        for(int j=0;j!=NVars;j++)
            BestInd[j] = Popul[besti][j];
    }
    if(bestfit < globalBestFit){
        globalBestFit = bestfit;
    }
}

void Optimizer::RemoveTooNear(int popSize, int NewNInds){
    int PointsToRemove = popSize - NewNInds;
    int curPopSize=popSize;
    
    double TOO_SMALL_DIST=1e-7;
    

    for(int L=0;L!=PointsToRemove;L++){
        double smallestDist=1e20;
        int best_i=0;
        int best_j=0;
        for(int i=0;i<curPopSize-1;++i){
            for(int j=i+1;j<curPopSize;++j){
                double dist= getDist(Popul[i], Popul[j], dim);
                if(dist<smallestDist){
                    smallestDist=dist;
                    best_i=i;
                    best_j=j;
                }
                
            }
        }
        if(smallestDist<=TOO_SMALL_DIST){
            for(int j=0;j!=NVars;++j)
                Popul[best_i][j] = Popul[curPopSize-1][j];
            Fitmass[best_i] = Fitmass[curPopSize-1];                
            --curPopSize;
        }

    }
    int pointsLeftToRemove = curPopSize - NewNInds;
    if(pointsLeftToRemove>0){
        RemoveWorst(curPopSize, NewNInds);
    }
}

void Optimizer::RemoveWorst(int popSize, int NewNInds)
{
    int PointsToRemove = popSize - NewNInds;
    for(int L=0;L!=PointsToRemove;L++)
    {
        double WorstFit = Fitmass[0];
        int WorstNum = 0;
        for(int i=1;i!=popSize;i++)
        {
            if(Fitmass[i] > WorstFit)
            {
                WorstFit = Fitmass[i];
                WorstNum = i;
            }
        }
        for(int i=WorstNum;i!=popSize-1;i++)
        {
            for(int j=0;j!=NVars;j++)
                Popul[i][j] = Popul[i+1][j];
            Fitmass[i] = Fitmass[i+1];
        }
    }
}

inline double Optimizer::GetValue(const int index, const int popSize, const int j)
{
    if(index < popSize)
        return Popul[index][j];
    return Archive[index-popSize][j];
}


void dumpInfo(ofstream& infoDump, int popSize, int Generation){
    infoDump<<popSize<<","<<Generation<<endl;
}

void dumpInd(ofstream& popDump, double* ind, const int dimensionality, double fit){
    for(int i=0;i<dimensionality; ++i){
        popDump<<ind[i]<<",";
    }
    popDump<<fit<<endl;
}

void Optimizer::MainCycle(double optFit){
    double ArchSuccess;
    double NoArchSuccess;
    double NArchUsages;
    double ArchProbs = arch_prob;
    
    double** centroid_matrix= new double*[popSize];
    for(int i=0;i<popSize;++i){
      centroid_matrix[i]=new double[dim];
    }


    vector<tuple<double, double, double>> abTab(dim);
    vector<double> centroids4dim(popSize);
    double* est_m = new double[dim]; 
    double* mean_indiv = new double[dim]; 
    double* mean_indivCl0 = new double[dim]; 
    double* mean_indivCl1 = new double[dim]; 
    double* mean_indiv_old = new double[dim]; 
    int* populLimCount = new int[popSize];
    for(int indIndx=0;indIndx<popSize;++indIndx){
        populLimCount[indIndx]=0;
    }
    int numOfStagIt=0;

    const int ITS_MODULO=17;

    double fit_mean_old=1e20;


    for(int curIndx=0;curIndx!=popSize;curIndx++)
    {

        Fitmass[curIndx] = cec_22_(Popul[curIndx],func_num, dim);

        FindNSaveBest(curIndx == 0, curIndx);
        if(!globalBestFitinit || bestfit < globalBestFit)
        {
            globalBestFit = bestfit;
            globalBestFitinit = true;
        }
        SaveBestValues(func_num, runNum);
    }
    do
    {
        double minfit = Fitmass[0];
        double maxfit = Fitmass[0];
        for(int i=0;i!=popSize;i++)
        {
            FitmassCopy[i] = Fitmass[i];
            Indexes[i] = i;
            if(Fitmass[i] >= maxfit)
                maxfit = Fitmass[i];
            if(Fitmass[i] <= minfit)
                minfit = Fitmass[i];
        }
        if(minfit != maxfit)
            qSort2int(FitmassCopy,Indexes,0,popSize-1);
        for(int i=0;i!=popSize;i++)
            for(int j=0;j!=popSize;j++)
                if(i == Indexes[j])
                {
                    BackIndexes[i] = j;
                    break;
                }
        FitTemp3.resize(popSize);
        for(int i=0;i!=popSize;i++)
            FitTemp3[i] = exp(-double(i)/(double)popSize);
        std::discrete_distribution<int> ComponentSelector3(FitTemp3.begin(),FitTemp3.end());

        double perTmp = per_of_the_best_considered_at_end-per_of_the_best_considered_at_start;
        if(perTmp<0)
            perTmp=0;
        int psizeval = max(2.0,popSize*(perTmp/(double)(maxFES)*(double)(NFEval)+per_of_the_best_considered_at_start));

        int CrossExponential = 0;
        if(Random(0,1) < prob_use_exp)
            CrossExponential = 1;
        for(int curIndx=0;curIndx!=popSize;curIndx++)
        {
            MemoryCurrentIndex = intRandom(MemorySize);
            Cr = min(1.0,max(0.0,NormRand(MemoryCr[MemoryCurrentIndex],norm_sigma)));
            do
            {
                F = CachyRand(MemoryF[MemoryCurrentIndex],cauchy_sigma);
            }
            while(F <= 0);
            FGenerated[curIndx] = min(F,1.0);
            CrGenerated[curIndx] = Cr;
        }
        qSort1(CrGenerated,0,popSize-1);
        double iterBestFit = 1E18;
        int iterBestIndx=0;
              
        for(int curIndx=0;curIndx!=popSize;curIndx++)
        {
            Rands[0] = Indexes[intRandom(psizeval)];
            for(int i=0;i!=25 && !CheckGenerated(0,Rands,curIndx);i++)
                Rands[0] = Indexes[intRandom(psizeval)];
            GenerateNextRandUnif(1,popSize,Rands,curIndx);
            if(Random(0,1) > ArchProbs || CurrentArchiveSize == 0)
            {
                Rands[2] = Indexes[ComponentSelector3(generator)];
                for(int i=0;i!=25 && !CheckGenerated(2,Rands,curIndx);i++)
                    Rands[2] = Indexes[ComponentSelector3(generator)];
                ArchUsages[curIndx] = 0;
            }
            else
            {
                GenerateNextRandUnifOnlyArch(2,popSize,CurrentArchiveSize,Rands,curIndx);
                ArchUsages[curIndx] = 1;
            }
            for(int j=0;j!=NVars;j++)
                Donor[j] = Popul[curIndx][j] +
                    FGenerated[curIndx]*(GetValue(Rands[0],popSize,j) - Popul[curIndx][j]) +
                    FGenerated[curIndx]*(GetValue(Rands[1],popSize,j) - GetValue(Rands[2],popSize,j));

            F = FGenerated[curIndx];//Corrected F
            int WillCrossover = intRandom(NVars);
            Cr = CrGenerated[BackIndexes[curIndx]];
            double CrToUse = 0;

            if(NFEval > when_start_2_use_CRtoUse*maxFES )
                CrToUse = (double(NFEval)/double(maxFES)-when_start_2_use_CRtoUse)*(1/(1-when_start_2_use_CRtoUse));

            if(CrossExponential == 0)
            {
                for(int j=0;j!=NVars;j++)
                {
                    if(Random(0,1) < CrToUse || WillCrossover == j)
                        populTemp[curIndx][j] = Donor[j];
                    else
                        populTemp[curIndx][j] = Popul[curIndx][j];
                }
            }
            else
            {
                int StartLoc = intRandom(NVars);
                int L = StartLoc+1;
                while(Random(0,1) < Cr && L < NVars)
                    L++;
                for(int j=0;j!=NVars;j++)
                    populTemp[curIndx][j] = Popul[curIndx][j];
                for(int j=StartLoc;j!=L;j++)
                    populTemp[curIndx][j] = Donor[j];
            }
            if(resampling){
                const int MAX_NUM_OF_TRIALS=100;
                int num_of_trials=1;
                bool usedRepair=false;
                if(IsInfeasible(populTemp[curIndx],NVars,Left,Right)){
                    usedRepair=true;
                    const int NUM_OF_TR_WITHOUT_F_HANGE=10;
                    do{
                        if(num_of_trials>NUM_OF_TR_WITHOUT_F_HANGE){
                            CrossExponential = 0;
                            if(Random(0,1) < prob_use_exp)
                                CrossExponential = 1;
                    
                    
                            MemoryCurrentIndex = intRandom(MemorySize);
                            Cr = min(1.0,max(0.0,NormRand(MemoryCr[MemoryCurrentIndex],norm_sigma)));
                            do{
                                F = CachyRand(MemoryF[MemoryCurrentIndex],cauchy_sigma);
                            }
                            while(F <= 0);
                            FGenerated[curIndx] = min(F,1.0);
                            CrGenerated[curIndx] = Cr;
                        }


                    Rands[0] = Indexes[intRandom(psizeval)];
                    for(int i=0;i!=25 && !CheckGenerated(0,Rands,curIndx);i++)
                        Rands[0] = Indexes[intRandom(psizeval)];
                    GenerateNextRandUnif(1,popSize,Rands,curIndx);
                    if(Random(0,1) > ArchProbs || CurrentArchiveSize == 0)
                    {
                        Rands[2] = Indexes[ComponentSelector3(generator)];
                        for(int i=0;i!=25 && !CheckGenerated(2,Rands,curIndx);i++)
                            Rands[2] = Indexes[ComponentSelector3(generator)];
                        ArchUsages[curIndx] = 0;
                    }
                    else
                    {
                        GenerateNextRandUnifOnlyArch(2,popSize,CurrentArchiveSize,Rands,curIndx);
                        ArchUsages[curIndx] = 1;
                    }
                    for(int j=0;j!=NVars;j++)
                        Donor[j] = Popul[curIndx][j] +
                            FGenerated[curIndx]*(GetValue(Rands[0],popSize,j) - Popul[curIndx][j]) +
                            FGenerated[curIndx]*(GetValue(Rands[1],popSize,j) - GetValue(Rands[2],popSize,j));
                    F = FGenerated[curIndx]; //corrected F
                    int WillCrossover = intRandom(NVars);
                    if(num_of_trials<=NUM_OF_TR_WITHOUT_F_HANGE){
                        Cr = CrGenerated[BackIndexes[curIndx]];
                    }
                    double CrToUse = 0;
                    if(NFEval > when_start_2_use_CRtoUse*maxFES )
                        CrToUse = (double(NFEval)/double(maxFES)-when_start_2_use_CRtoUse)*(1/(1-when_start_2_use_CRtoUse));
                    if(CrossExponential == 0)
                    {
                        for(int j=0;j!=NVars;j++)
                        {
                            if(Random(0,1) < CrToUse || WillCrossover == j)
                                populTemp[curIndx][j] = Donor[j];
                            else
                                populTemp[curIndx][j] = Popul[curIndx][j];
                        }
                    }
                    else
                    {
                        int StartLoc = intRandom(NVars);
                        int L = StartLoc+1;
                        while(Random(0,1) < Cr && L < NVars)
                            L++;
                        for(int j=0;j!=NVars;j++)
                            populTemp[curIndx][j] = Popul[curIndx][j];
                        for(int j=StartLoc;j!=L;j++)
                            populTemp[curIndx][j] = Donor[j];
                    }
                    ++num_of_trials;
                    }while(IsInfeasible(populTemp[curIndx],NVars,Left,Right)&&num_of_trials<=MAX_NUM_OF_TRIALS);
                    if(IsInfeasible(populTemp[curIndx],NVars,Left,Right)){
                        usedRepair=false;
                        FindLimits(populTemp[curIndx],Popul[curIndx],NVars,Left,Right);
                    }
                }

                if(count_lim){
                    const int MIN_ITS_ON_BOUND=this->min_its_on_bounds;
                    countLimits(populTemp[curIndx], &(populLimCount[curIndx]), NVars,Left,Right);
                    if(populLimCount[curIndx]>MIN_ITS_ON_BOUND){
                        cout<<"Bounds restart, curBestSol:"<<globalBestFit-optFit<<" FES:"<<NFEval<<"-----------------------------------------------"<<endl;
                        return;
                    }
                }
            }//if(resampling){
            popFitTmp[curIndx] = cec_22_(populTemp[curIndx],func_num, dim);
            if(popFitTmp[curIndx] < iterBestFit){
              iterBestFit = popFitTmp[curIndx];
              iterBestIndx=curIndx;
            }
            if(popFitTmp[curIndx] <= globalBestFit){
                globalBestFit = popFitTmp[curIndx];
                if(globalBestFit-optFit<=MIN_ERROR){
                  foundAt = NFEval;
                  ResultsArray[runNum][NUM_OF_DUMPS] = foundAt;
                  SaveBestValues(func_num, runNum);
                  cout<<"Found Opt----------------------------"<<endl;
                  return;
                }
            }

            if(popFitTmp[curIndx] < Fitmass[curIndx])
                SaveSuccessCrF(Cr,F,fabs(Fitmass[curIndx]-popFitTmp[curIndx]));
            
            FindNSaveBest(false,curIndx);
            SaveBestValues(func_num, runNum);
            
            if(NFEval>maxFES){
                cout<<"Max Evals----------------------------"<<endl;

                return;
            }            
        }
        int CHOSEN_INDX=0;

        if(k_means){
            arma::mat data(  dim, popSize ); // n_rows, n_cols
            // The assignments will be stored in this vector.
            for(int i=0;i<popSize;++i){
                for(int j=0;j<dim;++j){
                    data(j,i) = populTemp[i][j];//(i,j) at the i-th row and j-th column
                }
            }
            arma::Row<size_t> assignments;
            arma::Row<size_t> bestAssignments;
            arma::mat centroids;
            arma::mat bestCentroids;
            
            const int MAX_K=2;

            const int MIN_POP_SIZE_4_SPLIT = 20;
            double bestSilhouette=-1;
            int best_k=2;
            
            mlpack::metric::EuclideanDistance metric;
            if( popSize>=MIN_POP_SIZE_4_SPLIT){
                for( int cand_k=2; cand_k<=MAX_K; ++cand_k){
                    KMeans<> k;
                    k.Cluster(data, cand_k, assignments, centroids);    
                
                    double silhouetteScore = SilhouetteScore::Overall(data, assignments, metric);
                    if(silhouetteScore>bestSilhouette){
                        bestSilhouette=silhouetteScore;
                        best_k = cand_k;
                        bestAssignments = assignments;
                        bestCentroids = centroids;
                    }
                }
            }
            double bestCandFit=1e20;
            double MIN_silhouette = this->min_silhouette_mult*sqrt(dim);
            if(bestSilhouette>MIN_silhouette && popSize>=MIN_POP_SIZE_4_SPLIT){
                for( int cur_k=0; cur_k<best_k; ++cur_k){
                    for(int i=0;i<dim;++i){
                        mean_indiv[i] = bestCentroids(i,cur_k);
                        FindLimits(mean_indiv,mean_indiv,NVars,Left,Right);
                    }
                    double fit_mean = cec_22_(mean_indiv, func_num, dim); 

                    if(fit_mean<bestCandFit){
                        bestCandFit=fit_mean;
                        for(int i=0;i<dim;++i){
                            mean_indivCl0[i]=mean_indiv[i];
                        }
                    }

                    if(fit_mean < globalBestFit){
                        globalBestFit = fit_mean;
                        //memcpy(bestSol, mean_indiv, dimensionality*sizeof(double));
                    }        
                    SaveBestValues(func_num, runNum);
                    if(globalBestFit-optFit<=MIN_ERROR){
                        foundAt = NFEval;
                        ResultsArray[runNum][NUM_OF_DUMPS] = foundAt;
                        cout<<"Found Opt Mean----------------------------"<<endl;
                        return;
                    }

                    //Sale All in pop:
                    CHOSEN_INDX=getIndxOfNearest(populTemp, popSize, dim, mean_indiv);
                    if( fit_mean<popFitTmp[CHOSEN_INDX] ){
                        popFitTmp[CHOSEN_INDX] = fit_mean;
                        memcpy(populTemp[CHOSEN_INDX], mean_indiv, dim*sizeof(double));
                    }


                }//for( int cur_k=0; cur_k<best_k; ++cur_k){


            }else{//kmaeans failed- use just mean
                getMean(populTemp, popSize, dim, mean_indiv);
                FindLimits(mean_indiv,mean_indiv,NVars,Left,Right);
                double fit_mean=cec_22_(mean_indiv,func_num, dim);  
                if(fit_mean < globalBestFit){
                    globalBestFit = fit_mean;
                }        
                SaveBestValues(func_num, runNum);
                if(globalBestFit-optFit<=MIN_ERROR){
                    foundAt = NFEval;
                    ResultsArray[runNum][NUM_OF_DUMPS] = foundAt;
                    cout<<"Found Opt Mean----------------------------"<<endl;
                    return;
                }
                CHOSEN_INDX=getIndxOfNearest(populTemp, popSize, dim, mean_indiv);
                if( fit_mean<popFitTmp[CHOSEN_INDX] ){
                    popFitTmp[CHOSEN_INDX] = fit_mean;
                    memcpy(populTemp[CHOSEN_INDX], mean_indiv, dim*sizeof(double));
                }
            }//else{//kmaeans failed- use just mean
        }// end kmeans

        const double MIN_DIST=1e-9;
        const int MIN_numOfStagIt=this->min_num_stag_it;
        getMean(populTemp, popSize, dim, mean_indiv);
       if(getDist(mean_indiv_old, mean_indiv, dim)<MIN_DIST){
            numOfStagIt+=1;
            if(numOfStagIt>MIN_numOfStagIt){
                cout<<"Restart dist stagIt, curBestSol:"<<globalBestFit-optFit<<" FES:"<<NFEval<<"-----------------------------------------------"<<endl;
                return;
            }
        }else{
            numOfStagIt=0;
        }
        memcpy(mean_indiv_old, mean_indiv, dim*sizeof(double));

        ArchSuccess = 0;
        NoArchSuccess = 0;
        NArchUsages = 0;
        for(int curIndx=0;curIndx!=popSize;curIndx++)
        {
            if(popFitTmp[curIndx] <= Fitmass[curIndx])
            {
                if(ArchUsages[curIndx] == 1)
                {
                    ArchSuccess += (Fitmass[curIndx] - popFitTmp[curIndx])/Fitmass[curIndx];
                    NArchUsages += 1;
                }
                else
                    NoArchSuccess+=(Fitmass[curIndx] - popFitTmp[curIndx])/Fitmass[curIndx];
                CopyToArchive(Popul[curIndx]);
            }
            if(popFitTmp[curIndx] <= Fitmass[curIndx])
            {
                for(int j=0;j!=NVars;j++)
                    Popul[curIndx][j] = populTemp[curIndx][j];
                Fitmass[curIndx] = popFitTmp[curIndx];
            }
        }
        if(NArchUsages != 0)
        {
            ArchSuccess = ArchSuccess/NArchUsages;
            NoArchSuccess = NoArchSuccess/(popSize-NArchUsages);
            ArchProbs = ArchSuccess/(ArchSuccess + NoArchSuccess);
            ArchProbs = max(0.1,min(0.9,ArchProbs));
            if(ArchSuccess == 0)
                ArchProbs = arch_prob;
        }
        else
            ArchProbs = arch_prob;
        int newNInds;
        if(isRestart){//HomoAtReset
            double shapeConst = this->shape_const;
            const int MIN_POP=this->min_pop_size_aft_restart;


            double divider = (maxFES-evalsAtStart)/shapeConst;

            double delta=pow( (MIN_POP* (maxFES-evalsAtStart)/divider-(maxFES-evalsAtStart)/divider*popSize), 2)-4*(maxFES-evalsAtStart)/divider*(MIN_POP-popSize);
            if(delta<=1e-10){
                newNInds=MIN_POP;
            }else{
            double b1=(-(MIN_POP*(maxFES-evalsAtStart)/divider-(maxFES-evalsAtStart)/divider*popSize)-sqrt(delta))/(2*(MIN_POP-popSize));
            double a1=popSize-1/b1;
            newNInds = round(a1+1/( (NFEval-evalsAtStart) /divider+b1));
            }
        }else{
            newNInds = round((NIndsMin-NIndsMax)*pow((double(NFEval)/double(maxFES)),(1.0-double(NFEval)/double(maxFES)))+NIndsMax);
        }


        if(newNInds < NIndsMin)
            newNInds = NIndsMin;
        if(newNInds > NIndsMax)
            newNInds = NIndsMax;
        int newArchSize = round((NIndsMin-NIndsMax)*pow((double(NFEval)/double(maxFES)),(1.0-double(NFEval)/double(maxFES)))+NIndsMax)*ArchiveSizeParam;
        if(newArchSize < NIndsMin)
            newArchSize = NIndsMin;
        ArchiveSize = newArchSize;
        if(CurrentArchiveSize >= ArchiveSize)
            CurrentArchiveSize = ArchiveSize;
        RemoveWorst(popSize,newNInds);
        popSize = newNInds;
        UpdateMemoryCrF();
        SuccessFilled = 0;
        Generation ++;
    } while(NFEval < maxFES);
    delete [] est_m;
    delete [] mean_indiv;
}

void Optimizer::Clean()
{
    delete Donor;
    delete Trial;
    delete Rands;
    for(int i=0;i!=NIndsMax;i++)
    {
        delete Popul[i];
        delete populTemp[i];
    }
    for(int i=0;i!=NIndsMax*Int_ArchiveSizeParam;i++)
        delete Archive[i];
    delete ArchUsages;
    delete Archive;
    delete Popul;
    delete populTemp;
    delete Fitmass;
    delete popFitTmp;
    delete FitmassCopy;
    delete BestInd;
    delete Indexes;
    delete BackIndexes;
    delete tempSuccessCr;
    delete tempSuccessF;
    delete FGenerated;
    delete CrGenerated;
    delete FitDelta;
    delete MemoryCr;
    delete MemoryF;
    delete Weights;
}


int main(int argc, char* argv[]){
    po::options_description desc("Allowed options");
    
    double pop_size_mult=0.0;
    int k_means=1;
    int resampling=1;
    int pop_size_reset=0;
    double shape_const=0.0;
    int min_pop_size_aft_restart=0;
    int count_lim=0;
    int min_num_stag_it=0;
    double min_silhouette_mult=0.0;
    unsigned int seed;
    int min_its_on_bounds=0;

    int min_pop_size=4;
    int mem_size_mult=20;
    float arch_size_mult=2.1;
    float initial_CR=0.2;
    float initial_F=0.2;
    float arch_prob = 0.5;
    float prob_use_exp=0.5;
    float when_start_2_use_CRtoUse=0.5;
    float per_of_the_best_considered_at_start=20;
    float per_of_the_best_considered_at_end=60;
    float norm_sigma=0.1;
    float cauchy_sigma=0.1;
    float default_CR=0.5;
    float default_F=0.5;    
    std::string res_file_name;

    desc.add_options ()
        ("help", "print usage message")
        ("seed", po::value(&seed)->required(), "seed")
        ("pop_size_mult", po::value(&pop_size_mult)->required(), "pop_size_multiplier")
        ("k_means", po::value(&k_means)->required(), "use k_means")
        ("resampling", po::value(&resampling)->required(), "use resampling")
        ("pop_size_reset", po::value(&pop_size_reset)->required(), "pop_size after reset")
        ("shape_const", po::value(&shape_const)->required(), "shape_const")
        ("min_pop_size_aft_restart", po::value(&min_pop_size_aft_restart)->required(), "min_pop_size after restart")
        ("count_lim", po::value(&count_lim), "max it on bounds")
        ("min_num_stag_it", po::value(&min_num_stag_it), "min num of stagnated it to reset")
        ("min_silhouette_mult", po::value(&min_silhouette_mult), "min_silhouette_mult")
        ("min_its_on_bounds", po::value(&min_its_on_bounds), "min num of iterations  with given dim is on bound")
        ("mem_size_mult", po::value(&mem_size_mult), "memory size multiplier")
        ("arch_size_mult", po::value(&arch_size_mult), "archive size multiplier")
        ("initial_CR", po::value(&initial_CR), "initial_CR")
        ("initial_F", po::value(&initial_F), "initial_F")
        ("min_pop_size", po::value(&min_pop_size), "min_pop_size")
        ("arch_prob", po::value(&arch_prob), "prob of using archive")
        ("prob_use_exp", po::value(&prob_use_exp), "prob of using exp crossover")
        ("when_start_2_use_CRtoUse", po::value(&when_start_2_use_CRtoUse), "when_(fes) start_2_use_CRtoUse when CR is not Exp")
        ("per_of_the_best_considered_at_start", po::value(&per_of_the_best_considered_at_start), "per_of_the_best_considered_at_start")
        ("per_of_the_best_considered_at_end", po::value(&per_of_the_best_considered_at_end), "per_of_the_best_considered_at_end")
        ("norm_sigma", po::value(&norm_sigma), "norm_sigma")
        ("cauchy_sigma", po::value(&cauchy_sigma), "cauchy_sigma")
        ("default_CR", po::value(&default_CR), "default_CR")
        ("default_F", po::value(&default_F), "default_F")
        ("res_file_name", po::value<std::string>(&res_file_name)->required(), "result file base name")    

        ; 
    const int MAX_FUN_NMBR=12;
    int dims[]={10,20};
    // Parse command line arguments
    po::variables_map vm;
    po::store (po::command_line_parser (argc, argv).options (desc).run (), vm);
    if(vm.count("help")){
        std::cout << desc << "\n";
        return 0;
    }
    po::notify(vm);
    per_of_the_best_considered_at_start/=100;
    per_of_the_best_considered_at_end/=100;
    generator.seed(seed);
    double avResToRet=0;
    int MAX_DIMS=sizeof(dims)/sizeof(dims[0]);
    for (int dimIndx = 0;dimIndx<MAX_DIMS;++dimIndx){
        int dim=dims[dimIndx];
        if(dim == 10){
            maxFES = 200000;
        }
        if(dim == 20){
            maxFES = 1000000;
        }
        
        for (int fun_nmbr = 1;fun_nmbr<=MAX_FUN_NMBR;++fun_nmbr){
            int func_num = fun_nmbr;
            fopt = 0;

            switch(func_num)
            {
                case 1: {fopt = 300;  break;}
                case 2: {fopt = 400; break;}
                case 3: {fopt = 600;  break;}
                case 4: {fopt = 800; break;}
                case 5: {fopt = 900; break;}
                case 6: {fopt = 1800; break;}
                case 7: {fopt = 2000; break;}
                case 8: {fopt = 2200; break;}
                case 9: {fopt = 2300; break;}
                case 10:{fopt = 2400; break;}
                case 11:{fopt = 2600; break;}
                case 12:{fopt = 2700; break;}
            }
            unsigned optFound = 0;
            unsigned usedFes = 0;
            unsigned thresFound = 0;
            double NEXT_LEVEL_WEIGHT = 0.01;
            const unsigned lengthOut=51;
            double thresholds[lengthOut];
            double seqStart = 3; 
            double seqStop = -8;
            double curSeq=seqStart;
            double seqDelta = (seqStart - seqStop)/(lengthOut);
            for(int i=0; i<lengthOut;++i){
                thresholds[i] = pow(10, curSeq);
                curSeq-=seqDelta;
            } 
            int POP_SIZE=0;
            for (int runNum = 0;runNum!=NMBR_OF_RUNS;++runNum){
                POP_SIZE=dim*pop_size_mult;

                string startPopFilename="../startPointsGen/startingPoints";
                startPopFilename+=to_string(dim)+'_'+to_string(runNum)+'_'+to_string(fun_nmbr)+".txt";
                ifstream fin(startPopFilename.c_str());
                vector<vector<double>> startPop;
                if(fin.is_open()){
                    double f;
                    char c;
                    for(int numOfStartPoints = 0;numOfStartPoints<POP_SIZE; ++numOfStartPoints){
                        vector<double> startInd;
                        for(int curDim = 0;curDim<dim; ++curDim){
                            fin>>f;
                            fin>>c;
                            startInd.push_back(f);
                        }
                        startPop.emplace_back(startInd);
                    }
                }else{
                    cout<<startPopFilename<<" open failed"<<endl;
                    return -1;
                }                

                int oldPopSize=POP_SIZE;

                globalBestFitinit = false;
                initfinished = false;
                LastFEcount = 0;
                NFEval = 0;
                Optimizer OptZ;
                ////////////////popSize   NVars  func     Run  bench     memory    arch size 
                OptZ.Initialize(POP_SIZE, dim, func_num, mem_size_mult*dim, arch_size_mult, k_means, resampling, shape_const, min_pop_size_aft_restart, count_lim, min_num_stag_it, min_silhouette_mult, min_its_on_bounds,
                    initial_CR, initial_F, min_pop_size, arch_prob, prob_use_exp, when_start_2_use_CRtoUse, per_of_the_best_considered_at_start, per_of_the_best_considered_at_end, 
                    norm_sigma, cauchy_sigma, default_CR, default_F, runNum, startPop);
                isRestart=false;
                do{
                    OptZ.MainCycle(fopt);
                    POP_SIZE = pop_size_reset; //best
                    OptZ.restart(POP_SIZE);
                    isRestart=true;

                }while(globalBestFit-fopt>MIN_ERROR && NFEval<maxFES);
                POP_SIZE=oldPopSize;
                isRestart=false;

                double fitError=globalBestFit-fopt;
                if(fitError<=1e-8){
                    ++optFound;
                    usedFes += NFEval;
                    thresFound+=lengthOut;
                }else{
                    usedFes += maxFES;
                    for(int i=0; i<lengthOut;++i){
                        if(fitError<=thresholds[i]){
                            ++thresFound;
                        }else{
                            break;
                        }
                    }
                }
                OptZ.Clean();
            }
            string cur_res_file_name = res_file_name+"_" + to_string(fun_nmbr)+"_"+to_string(dim)+ ".txt";
            ofstream resArrayFile(cur_res_file_name.c_str(), ios::trunc);
            for(int step = 0;step!=NUM_OF_DUMPS+1;++step){
                for(int runNum = 0;runNum!=NMBR_OF_RUNS;++runNum){
                    if(step==NUM_OF_DUMPS && ResultsArray[runNum][step]==0){
                        ResultsArray[runNum][step] = maxFES;
                    }
                    resArrayFile<<ResultsArray[runNum][step]<<"\t";
                }
                    
                resArrayFile<<endl;
            }

            double foundToRet = 100.0 * optFound/NMBR_OF_RUNS;
            double thresToRet = 100.0 * thresFound/(NMBR_OF_RUNS*lengthOut);
            double evalsLeftToRet = 100.0 - 100.0 * usedFes/(maxFES*NMBR_OF_RUNS);
            double resToRet = foundToRet+thresToRet*NEXT_LEVEL_WEIGHT+evalsLeftToRet*NEXT_LEVEL_WEIGHT*NEXT_LEVEL_WEIGHT;
            avResToRet += resToRet;
        }//foreach fun
    }//foreach dim
    cout<<-avResToRet/(MAX_DIMS*MAX_FUN_NMBR)<<endl;
 
        
	return 0;
}


int intRandom(int target){
    if(target == 0)
        return 0;
    return uni_int(generator)%target;
}
